import socket
import threading
import datetime
import os
import argparse


def handle_tcp_client(client_socket, address, output_dir):
    """Handles a single TCP client connection."""
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")
    filename = os.path.join(
        output_dir, f"tcp_{address[0]}_{address[1]}_{timestamp}.bin"
    )
    print(f"TCP: Handling connection from {address}.  Saving to {filename}")

    try:
        with open(filename, "wb") as f:
            while True:
                data = client_socket.recv(4096)
                if not data:
                    break
                f.write(data)
        print(f"TCP: Connection from {address} closed. Data saved.")
    except Exception as e:
        print(f"TCP: Error handling connection from {address}: {e}")
    finally:
        client_socket.close()


def tcp_server(port, output_dir):
    """Starts a TCP server on a specific port."""
    try:
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(("", port))
        server_socket.listen(5)
        print(f"TCP server listening on port {port}")

        while True:
            client_socket, address = server_socket.accept()
            client_thread = threading.Thread(
                target=handle_tcp_client, args=(client_socket, address, output_dir)
            )
            client_thread.start()
    except Exception as e:
        print(f"Error starting TCP server on port {port}: {e}")
        return
    finally:
        if "server_socket" in locals():
            server_socket.close()


def handle_udp_packet(data, address, output_dir, port):
    """Handles a single UDP packet."""
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")
    filename = os.path.join(
        output_dir, f"udp_{address[0]}_{address[1]}_{port}_{timestamp}.bin"
    )  # Include port.
    print(f"UDP: Received packet from {address} on port {port}. Saving to {filename}")

    try:
        with open(filename, "wb") as f:
            f.write(data)
        print(f"UDP: Packet from {address} saved.")
    except Exception as e:
        print(f"UDP: Error handling packet from {address}: {e}")


def udp_server(port, output_dir):
    """Starts a UDP server on a specific port."""
    try:
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        server_socket.bind(("", port))
        print(f"UDP server listening on port {port}")

        while True:
            data, address = server_socket.recvfrom(4096)
            udp_thread = threading.Thread(
                target=handle_udp_packet, args=(data, address, output_dir, port)
            )
            udp_thread.start()
    except Exception as e:
        print(f"Error starting UDP server on port {port}: {e}")
    finally:
        if "server_socket" in locals():
            server_socket.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="TCP/UDP server to dump received data to files."
    )
    parser.add_argument(
        "-t",
        "--tcp_ports",
        type=str,
        default="",
        help="Comma-separated list of TCP ports",
    )
    parser.add_argument(
        "-u",
        "--udp_ports",
        type=str,
        default="",
        help="Comma-separated list of UDP ports",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Output directory (default: current directory)",
    )

    args = parser.parse_args()

    os.makedirs(args.output, exist_ok=True)

    if not args.tcp_ports and not args.udp_ports:
        print("Please specify at least one TCP or UDP port list.")
        exit(1)

    tcp_ports = []
    if args.tcp_ports:
        try:
            tcp_ports = [int(port) for port in args.tcp_ports.split(",")]
        except ValueError:
            print("Invalid TCP port list.  Must be comma-separated integers.")
            exit(1)

    udp_ports = []
    if args.udp_ports:
        try:
            udp_ports = [int(port) for port in args.udp_ports.split(",")]
        except ValueError:
            print("Invalid UDP port list.  Must be comma-separated integers.")
            exit(1)

    threads = []

    for port in tcp_ports:
        tcp_thread = threading.Thread(target=tcp_server, args=(port, args.output))
        tcp_thread.start()
        threads.append(tcp_thread)

    for port in udp_ports:
        udp_thread = threading.Thread(target=udp_server, args=(port, args.output))
        udp_thread.start()
        threads.append(udp_thread)

    for t in threads:
        t.join()
